/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#include "qubopt.h"

#include "QUB_FitCurves.h"
using namespace std;


#define QUB_FITFREEZER_MAX 1000000


void QUB_FitCurve_impl::setupStorage(int lvl)
{
	switch ( lvl )
	{
	case 0:
		data.loadData();
		storage |= QUB_FS_0;
		break;
	case 1:
		storage |= QUB_FS_1;
		break;
	}
}

void QUB_FitCurve_impl::clearStorage(int lvl)
{
	switch ( lvl )
	{
	case 0:
		data.unloadData( changed );
		changed = false;
		storage &= ~ QUB_FS_0;
		break;
	case 1:
		storage &= ~ QUB_FS_1;
		break;
	}
}



QUB_FitCurveIter_impl::QUB_FitCurveIter_impl( QUB_FitCurves_impl& curvs, int f )
: curves( curvs ), fseries( curvs.fseries ), nodes( curvs.nodes ), locations( curvs.locations ), changed( false )
{
	curves.lock();

	// lower_bound(x) returns first elt with key >= x
	locIter = locations.lower_bound( f );
	if ( locIter != locations.end() )
		curve = fseries.checkOut( locIter->second, QUB_FS_0 | QUB_FS_1 );
	else
		curve = NULL;

	while ( 1 ) {
		if ( curve && (curve->Last() < f) ) {
			fseries.checkIn( locIter->second, false );
			++locIter;
			if ( locIter != locations.end() )
				curve = fseries.checkOut( locIter->second, QUB_FS_0 | QUB_FS_1 );
			else
				curve = NULL;
			break;
		}
		if ( locIter == locations.begin() || (curve && curve->First() == f ))
			break;
		if ( curve )
			fseries.checkIn( locIter->second, false );
		--locIter;
		if ( locIter != locations.end() )
			curve = fseries.checkOut( locIter->second, QUB_FS_0 | QUB_FS_1 );
		else
			curve = NULL;
	}

	if ( locIter != locations.end() )
		curve->setupStorage( 0 );
}

QUB_FitCurveIter_impl::~QUB_FitCurveIter_impl()
{
	if ( curve )
		fseries.checkIn(locIter->second, changed);

	curves.unlock();
}

int            __cdecl  QUB_FitCurveIter_impl::Next()
{
	if ( curve ) {
		fseries.checkIn(locIter->second, changed);
		++locIter;
		if ( locIter != locations.end() ) {
			curve = fseries.checkOut(locIter->second, QUB_FS_0 | QUB_FS_1);
			curve->setupStorage( 0 );
		}
		else
			curve = NULL;
	}

	changed = false;
	return ( locIter == locations.end() );
}



QUB_FitCurves_impl::QUB_FitCurves_impl()
: fseries( &freezer )
{
	InitializeCriticalSection(&crit);
	freezer.setThreshold(0, QUB_FITFREEZER_MAX);
}

QUB_FitCurves_impl::~QUB_FitCurves_impl()
{
	DeleteCriticalSection(&crit);
}

void              __cdecl  QUB_FitCurves_impl::Free()
{
	delete this;
}

void              __cdecl  QUB_FitCurves_impl::Clear()
{
	lock();
	fseries.truncate(0);
	nodes.resize(0);
	locations.clear();
	unlock();
}

QUB_FitCurveIter* __cdecl  QUB_FitCurves_impl::Add(int first, int last)
{
	lock();
	int ix = fseries.size();
	nodes.resize(ix+1);
	QUB_Tree data = nodes[ix];
	QUB_FitCurve_impl* curve = new QUB_FitCurve_impl(first, last, data);
	fseries.insert(ix, curve);
	locations[first] = ix;
	QUB_FitCurveIter* iter = Find(first);
	unlock();
	return iter;
}

QUB_FitCurveIter* __cdecl  QUB_FitCurves_impl::Find(int first)
{
	return new QUB_FitCurveIter_impl(*this, first);
}

int              __cdecl  QUB_FitCurves_impl::Empty()
{
	return ! fseries.size();
}

void              __cdecl  QUB_FitCurves_impl::lock()
{
	EnterCriticalSection(&crit);
}

void              __cdecl  QUB_FitCurves_impl::unlock()
{
	LeaveCriticalSection(&crit);
}

void              __cdecl  QUB_FitCurves_impl::UseTempFile()
{
	nodes.getRoot().saveAsTemp();
}

extern "C" MIL_API QUB_FitCurves*    __stdcall QUB_FitCurves_Create()
{
	return new QUB_FitCurves_impl;
}
